/*
   author:  Rony G. Flatscher
   date:    2005-04-18/19/20

   last change: $Revision: 178 $ $Author: rony $ $Date: 2006-11-19 13:39:27 +0100 (Sun, 19 Nov 2006) $

   changed:
            2006-09-03, ---rgf, added public routine "uno.getInterfaceName": returns fully qualified
                                name of unqualified interface name; if duplicates are possible a blank
                                delimited string of the fully qualified interface names is returned
            2006-07-24, ---rgf, made UNO_CONSTANTS.encode() a little bit more flexible (now accepting
                                an arbitrary mix of constant-field-names and decimal values)
            2006-07-19, ---rgf, added/changed class UNO_DIRLIKE, UNO_ENUM and UNO_CONSTANTS, which
                                can be treated as directories, if retrieving field-names or their
                                numeric values; UNO_CONSTANTS has in addition the methods
                                ENCODE(blank-delimited-field-names) returning the encoded decimal
                                number; DECODE(decimal number) returning a blank-delimited-filed-names
            2006-07-17, ---rgf, added UNO.WRAPARRAY which determines whether to wrap up a BSF_ARRAY
                                object as an UNO_ARRAY_PROXY
            2006-07-16, ---rgf, added UNO_ARRAY_PROXY class; the public routine UNO.CreateArray()
                                which uses BSF.CreateArray(), but wraps it up as an UNO_ARRAY_PROXY
                                (making sure that its elements are wrapped up as UNO_PROXY objects)
            2006-07-12, ---rgf, uno.setCell() now allows letter-number-coordinates; added routine
                                uno.getCell(xSheet, colNr0, rowNr0) or uno.getCell(xSheet, LetterNumber)
                                returning a XCell object [you can use {g|s}etValue(), {g|s}etFormula(),
                                getError() or getType() with it]
            2006-07-10, ---rgf, in uno.getScriptContext() the RgfReflectUNO class gets explicitly a XComponentContext
                                set to; otherwise RgfReflectUNO will bootstrap hanging the OOo Scripting
                                Framework for ooRexx macros dispatched by it!
                                Just for congruency, also added same technique to UNO.connect().
                                Now fallback strategy in UNO_PROXY::UNKNOWN re-established: if unknown
                                interface name then reflection is used at runtime
            2006-06-04, ---rgf, fixed a bug in querying an interface object, if unqualified class name is
                                used more than once
            2006-02-03, ---rgf, commented fall-back-strategy to query interface object via RgfReflectUNO-class
                                in UNO_PROXY.UNKNOWN
            2006-01-06, ---rgf, removed public routines pp() and iif(), because they are now availble
                                in the required BSF.CLS module; added the public routines
                                uno.getScriptContext() and uno.getScriptContextVersion()

            2005-05-03, ---rgf, made sure that exceptions are propagated
            2005-06-05, ---rgf, uppercasing "OOo.cls" to "OOO.CLS", removing
            2005-07-03, ---rgf, changed logic to queryInterface(), now dealing with duplicate class names
                                gracefully; fixed a bug (do not wrap up BSF arrays)
            2005-07-04, ---rgf, optimized handling interface classes,
                                added OOo.getInterfaces(object[, sortByUnqualifiedClassName]),
                                added OOo.getServices(object)
            2005-07-05, ---rgf, removed unnecessary preloading of application classes
            2005-07-08, ---rgf, ooo.wrap(): now tries a bsf.wrap(), if ncecessary
            2005-07-30, ---rgf, adapted OOO.CLS to new array-behaviour in BSF.CLS
            2005-08-18, ---rgf, added public routines "ConvertToURL()", "ConvertFromURL()", which use
                                "encodeURL()", "decodeURL()"
            2005-08-31, ---rgf, renamed module to "UNO.CLS" and renamed stuff from "OOO" to "UNO" to
                                reflect the generic support for UNO; office is just one combination of
                                UNO components and UNO application; added license
            2005-12-15, ---rgf, added newly developed Java class "org.oorexx.uno.RgfReflectUNO",
                                added entry "version" in ".uno"
            2005-12-16, ---rgf, added wrapper routines for RgfReflectUNO methods which wrap returned
                                UNO objects into ooRexx UNO_PROXY objects;
            2005-12-21, ---rgf, adapted and added wrapper routines for RgfReflectionUNO
            2005-12-26, ---rgf, adapted UNO-routine names to new RgfReflectionUNO names, added
                                UNO.*-methods to UNO_PROXY
            2005-12-27, ---rgf, updated license part, added utility class "UNO_ENUM_OR_CONSTANTS"
            2005-12-28, ---rgf, generalized the "directoryfiying" of UNOIDL definitions for "UNO_ENUM",
                                "UNO_CONSTANTS" and "UNO_PROPERTIES" (removed "UNO_ENUM_OR_CONSTANTS")
            2005-12-29, ---rgf, "UNO_ENUM" now returns the Java object representing the enum-value,
                                "UNO_CONSTANTS" will return the constant's value (a number)
            2006-01-03, ---rgf, corrected inline documentation


   version: 1.1.0
   name:    UNO.CLS
   purpose: allow interfacing with UNO components (via Java), tested with OpenOffice.org, v1.1.x and up
   needs:   BSF4Rexx, BSF.CLS, UNO_XINTERFACES.REX

   license:

    ------------------------ Apache Version 2.0 license -------------------------
       Copyright (C) 2006 Rony G. Flatscher

       Licensed under the Apache License, Version 2.0 (the "License");
       you may not use this file except in compliance with the License.
       You may obtain a copy of the License at

           http://www.apache.org/licenses/LICENSE-2.0

       Unless required by applicable law or agreed to in writing, software
       distributed under the License is distributed on an "AS IS" BASIS,
       WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
       See the License for the specific language governing permissions and
       limitations under the License.
    -----------------------------------------------------------------------------
*/


   -- if no environment entry for UNO as of yet, then create the entry with default classes
if .local~hasentry("UNO")=.false then
do
   .local~UNO=.directory~new     -- create a new directory object and store it in .local
   .uno~version="111.20060724"   -- set version number
   .uno~bExtendSearch=.true      -- as of 2005-12-20: if interface not found via XTypeProvider,
                                 -- use full reflection (slow, but there seem to be bugs in XTypeProvider)

   -- save class objects of UNO proxy classes defined in this module
   .uno~proxyclasses=.set~of(.uno_proxy, .uno_array_proxy)

   -- preload some important classes, just send their class name to the directory object ".UNO"
   l = .list~of( "com.sun.star.beans.PropertyValue"   , -
                 "org.oorexx.uno.RgfReflectUNO"       , -
                 "com.sun.star.comp.helper.Bootstrap" , -
                 "com.sun.star.uno.Any"               , -
                 "com.sun.star.uno.AnyConverter"      , -
                 "com.sun.star.uno.UnoRuntime"       )

   do c over l -- get Java class object and save it in .UNO by className
      call UNO.loadClass c
   end

-- trace ?all
-- trace ?i

      -- create empty array of property-value (needed quite often)
   .UNO~noProps=.bsf~bsf.createArray(.UNO~propertyValue, 0)

      -- list of Interface classes to be loaded at runtime (needs completion)
   l = UNO.getInterfaceList()    -- retuns > 950 UNO-interface classes in a list

   -- save interface class names with which we want to deal
   d=.directory~new  -- directory for all InterfaceTypes
   r=.relation~new   -- relation to store duplicates (if last word is used multiple times, keep full class name)
   do i over l
     idx=substr(i, lastpos(".", i)+1)
     d~setentry(i, i)    -- by fully qualified name

     if d~hasentry(idx) then  -- interface class name used already, use relation to store full name
     do
        r[idx~translate]=i    -- save full name with class name
     end
     else   -- new interface class name
     do
       d~setentry(idx, i)     -- by last word
     end
   end
   .UNO~XInterfaces = d       -- save directory
   .UNO~XInterfaces.dupes = r -- save relation
end


::requires BSF.CLS                -- get BSF4Rexx support
::requires UNO_XINTERFACES.REX    -- get the list of all of UNO XInterface classes


      -- load UNO-class and save it in environment; "idx" determines the name which is used to
      -- store the class object in the directory ".UNO"
::routine UNO.loadClass public
     parse arg className, idx

     if idx="" then idx=substr(className, lastpos(".", className)+1) -- extract last word to serve as index
     if .UNO~hasentry(idx)=.true then    -- e.g. last word is used as a class in a another package as well
        .error~say("---> UNO.loadClass():" pp(idx) "already used, in hand:" pp(className) "in .UNO:" pp(.UNO~entry(idx)))

     tmpClass=.UNO_proxy~new(.bsf~bsf.import(className, .nil))  -- make it look like an ooRexx class

     .UNO~setentry(idx, tmpClass)   -- load the given Java class and save it in .UNO
     return tmpClass


   /*
      connects to the UNO runtime environment and returns the XContext object; if passing a URL,
      then the remote object is returned
   */
::routine UNO.connect public
  parse arg unoURL

  xContext=.UNO~Bootstrap~bootstrap          -- get an initialized local UNO runtime environment

   -- make sure RgfReflectUNO() has XComponentContext available, so it does not need to bootstrap
   .uno~rgfReflectUno~setContext(xContext)

  if unoURL="" then  -- no unoURL given, use default (local installation)
  do
     return xContext       -- return the context object (which can be used to get the service manager)
  end

  uur= xContext~getServiceManager~createInstanceWithContext("com.sun.star.bridge.UnoUrlResolver", xContext)
  remoteObject = uur~XUnoUrlResolver ~resolve(unoURL)    -- get the connection
  return remoteObject

/*
 -- unoURL="uno:socket,host=localhost,port=8100;urp;StarOffice.ServiceManager"
    remoteMSF = remoteServiceManager~XMultiServiceFactory

--- or:
--  unoURL="uno:socket,host=localhost,port=8200;urp;StarOffice.NamingService"
  remoteNamingService = remoteObject~XNamingService

  remoteServiceManager=remoteNamingService~getRegisteredObject("StarOffice.ServiceManager")
  remoteMSF           = remoteServiceManager~XMultiServiceFactory
  if .UNO~hasentry("remoteMSF")=.false then .UNO~remoteMSF=remoteMSF -- save remote MultiServiceFactory
  return remoteMSF  -- return the remote XMultiServiceFactory
*/


      -- get a XContext, use it to create and to return the XDesktop interface object
::routine UNO.createDesktop public
  USE ARG xContext

  if arg()=0 | .nil=xContext then   -- no context given, create default (local) one
     xContext=UNO.connect()

  service="com.sun.star.frame.Desktop" -- OOo Desktop UNO component service
  return xContext~getServiceManager~createInstanceWithContext(service, xContext)


   -- a routine for inserting values or formulas into cells; uses setFormula() which is language independent
::routine UNO.setCell public

   if arg()=4 then   -- address is by position, directly addressing an individual cel
   do
      use arg xSheet, x, y, content
   	xSheet~getCellByPosition(x, y) ~setFormula(content)
      return
   end

   -- cell was addressed by a nameAdress (ie. a "cellRange"), e.g. "E4"
   use arg xSheet, nameAddress, content
      -- get the cellRange and address its first cell at position (0,0)
   xSheet~getCellRangeByName(nameAddress)~getCellByPosition(0,0) ~setFormula(content)


   -- a routine for getting and returning a XCell object; understands setValue(), getValue(),
   -- getFormula(), setFormula(), getError(), getType()
::routine UNO.getCell public

   if arg()=3 then   -- address is by position, directly addressing an individual cel
   do
      use arg xSheet, x, y
   	return xSheet~getCellByPosition(x, y)
   end

      -- cell was addressed by a nameAdress (ie. a "cellRange"), e.g. "E4"
   use arg xSheet, nameAddress
      -- get the cellRange and address its first cell at position (0,0)
   return xSheet~getCellRangeByName(nameAddress)~getCellByPosition(0,0)





/* *************************************************************
   Class:       UNO_PROXY
   Purpose:     allows to wrap up Java objects as Object Rexx objects;
                if instances are destroyed, then no derigstration from
                the BSF registry takes place
   restriction: needed to make getting interface objects easier
   ************************************************************* */
::CLASS UNO_PROXY -- subclass BSF PUBLIC

::Method bsfObject ATTRIBUTE

::METHOD init
   expose arg bsfObject -- direct access to BSF object
   use arg bsfObject    -- assign object to attribute

      -- set beanName
   self~objectName=bsfObject~objectName   -- use same objectname as BSF proxy (= idx into BSF-registry)

      -- if bsfObject is an imported Java class object, then use its beanName directly
   if bsfObject~string~left(4)~translate="THE " then
   do
      signal on syntax name continue   -- if an error, not a JCO, hence use its "plain" objectname
      self~objectname=.bsf~bsf.getJavaClassObjectProxy(bsfObject~bsf.FullJavaClassName)~string
   end
   return

continue:   -- catch exception
/*
   -- default objectname
   self~objectName=bsfObject~objectName   -- use same objectname as BSF proxy (= idx into BSF-registry)
say "UNO.CLS ---> ---> self~objectName="self~objectname
*/


::METHOD unknown
   expose arg bsfObject -- direct access to BSF object
   parse upper arg xName 1 idx   -- make sure message name is in uppercase (if quoted)

   if pos(".", idx)>0 then idx=substr(idx, lastpos(".", idx)+1)   -- getindex

   SIGNAL ON SYNTAX NAME ANY  -- intercept all exceptions
   if idx~left(1)="X" then    -- if an interface object is asked for, query and return it
   do
      xClass=getInterfaceClass(xName, idx)

         -- query and return the interface object
      interface=.UNO~unoRuntime~queryInterface(xClass, bsfObject)
      if interface=.nil then  -- not found, maybe a duplicate simple class name ?
      do
         do fullName over .UNO~XInterfaces.dupes~allAt(idx)
            idx=substr(fullName, lastpos(".", fullName)+1)   -- getindex
            xClass=getInterfaceClass(fullName, idx)      -- rgf, 2006-06-04

            if .nil<>xClass then
               interface=.UNO~unoRuntime~queryInterface(xClass, bsfObject)

            if .nil<>interface then
               return .UNO_proxy~new(interface)
         end

         -- now try reflection (via "RgfReflectUNO"), maybe a new interface, not in UNO_XINTERFACES.REX
         res=.uno~RgfReflectUno~queryInterfaceObjectByName(bsfObject, xName, .uno~bExtendSearch )

         if .nil<>res then
            return .uno_proxy~new(res)

         return .nil    -- no interface object available
      end

      return .UNO_proxy~new(interface)
   end
   else  -- pass to BSF to handle it
   do
      if bsfObject~hasmethod(xName) then -- o.k. it's a BSF method (could be ooRexx OBJECT method as well!)
         FORWARD TO (bsfObject) MESSAGE (xName) ARGUMENTS (arg(2))
      else
         FORWARD TO (bsfObject) CONTINUE -- let BSF handle this

      tmpObj=result     -- retrieve result value

      -- if a BSF-object, but not a BSF array: wrap it up
      if tmpObj~hasmethod("bsf.setFieldValue") then
      do
           -- if a BSF_ARRAY was returned wrap it up as an UNO_ARRAY_PROXY
         tmpClass=tmpObj~class~id   -- get class name
         if abbrev(tmpClass, "BSF_ARRAY") then
            return uno.wrapArray(tmpObj)

           -- plain BSF object, hence wrap it up
         return .UNO_proxy~new(tmpObj)
      end

      return tmpObj
   end

ANY:
  RAISE PROPAGATE    -- raise exception in caller

-- local function
-- try to get Interface class object; if not available, load it ("lazy loading")
getInterfaceClass: procedure
   parse arg xName, idx

   if xName<>idx then xClass=.UNO~entry(xName)  -- try to get interface class object, if already created
                 else xClass=.UNO~entry(idx)    -- try to get interface class object, if already created

   if xClass=.nil then        -- no entry as of yet
   do
      if xName<>idx then      -- long name given, then use that
      do
         xClass=UNO.loadClass(.UNO~XInterfaces~entry(xName), xName)
         if xClass<>.nil then -- o.k. also save with class name
         do
             .UNO~setEntry(idx, xClass)
         end
      end
      else     -- only class name is given
      do
         xClass=UNO.loadClass(.UNO~XInterfaces~entry(idx), idx)
         if xClass<>.nil then -- o.k., also save with full class name
         do
             .UNO~setEntry(.UNO~XInterfaces~entry(idx), xClass)
         end
      end
   end

   return xclass



   -- interface to UnoRuntime.areSame(o1, o2)
::method uno.isSame
   use arg o
   return .uno~UnoRuntime~areSame(self, o) /* true if both reference the same UNO object   */


-- supply the UNO.-routines in form of methods as well, makes it easier to use them

   /* interface to RgfReflectUNO.findInterfaceWithMember(): can be a string or an object
      if "bReturnString=.false", then carry out a queryInterface() and return that object (or
      .nil); if a string should be returned ("bReturnString=.true"), then "howMany" determines
      after how many matches the search should be ended (default "-1", ie. try to find all
      matches)
   */
::method uno.findInterfaceWithMember
   use arg name, bReturnString, howMany
   return uno.findInterfaceWithMember(self, name, bReturnString, howMany)


   -- interface to RgfReflectUNO.getDefinition(): returns always a blank delimited string
::method uno.getDefinition
   return uno.getDefinition(self)


   -- interface to RgfReflectUNO.getInterfacesViaReflection(): returns always a blank delimited string
::method uno.getInterfaceNamesViaReflection
   return uno.getInterfaceNamesViaReflection(self)


   -- interface to RgfReflectUNO.getProperties(): returns always a blank delimited string
::method uno.getProperties
   return uno.getProperties(self)


   -- interface to RgfReflectUNO.getServicesViaReflection(): returns always a blank delimited string
::method uno.getServiceNamesViaReflection
   return uno.getServiceNamesViaReflection(self)


   -- interface to RgfReflectUNO.getType(): returns always a string
::method uno.getTypeName
   return uno.getTypeName(self)


   -- interface to RgfReflectUNO.getXTypeProviderTypes(): returns always a blank delimited string
::method uno.getXTypeProviderTypeNames
   return uno.getXTypeProviderTypeNames(self)


   -- interface to RgfReflectUNO.queryInterfaceName(): returns always a string
::method uno.queryInterfaceName
   use arg name
   return uno.queryInterfaceName(self, name)


   -- interface to RgfReflectUNO.queryInterfaceObjectByName(): returns the interface object or .nil
::method uno.queryInterfaceObjectByName
   use arg name
   return uno.queryInterfaceObjectByName(self, name)


   -- interface to RgfReflectUNO.queryServiceName(): returns always a string
::method uno.queryServiceName
   use arg name
   return uno.queryServiceName(self, name)



/* *************************************************************
   Class:      UNO_ARRAY_PROXY
   Purpose:    allows to wrap up a BSF_ARRAY object; this way we can make sure that
               its elements are wrapped up as uno proxy objects
   ************************************************************* */
::CLASS UNO_ARRAY_PROXY SUBCLASS UNO_PROXY
::method "[]"                    -- intercept access to array element
  forward message ("AT")         -- let AT do the work

::method at                      -- make sure to return a UNO_PROXY, such that we can send interface names as messages
  forward class (super) continue -- let superclass carry out its work (will be directed to BSF.CLS)
  return uno.wrap(result)        -- try to wrap up object as an UNO object

::method makearray               -- iterate over all elements and wrap them up as UNO_objects?
  forward class (super) continue -- let superclass carry out its work (will be directed to BSF.CLS)
  arr=result
  do i=1 to arr~items            -- iterate over all entries
     arr[i]=uno.wrap(arr[i])     -- wrap up each entry as an UNO proxy
  end

  return arr




/* *************************************************************
   Class:      UNO_DIRLIKE
   Purpose:    allows to make all UNO definitions for ENUM, CONSTANTS and PROPERTIES
               available via an ooRexx directory object, easying lookups considerably
   ************************************************************* */
::CLASS  UNO_DIRLIKE -- SUBCLASS Relation -- Directory
::method unoidl_name       attribute   -- fully qualified name of the enum/constant groupe
::method unoidl_definition attribute   -- UNOIDL definition in form of a blank delimited string
::method unoidl_typeName   attribute   -- UNOIDL typeName, will be .nil, if setup could not be carried out
::method unoidl_className  attribute   -- the fully qualified Java class name representing

::method unoDirectory      attribute   -- maps UNO_NAME->numValue, numValue->UNO_NAME, objectName->UNO_NAME
::method unoNameQueue      attribute   -- collects UNO_NAMEs defined

::method init
       -- establish access to object attributes
    expose unoidl_name unoidl_definition unoidl_typeName unoidl_className unoDirectory unoNameQueue
    parse arg unoidl_name              -- parse argument as string

    unoidl_typeName=uno.getTypeName(unoidl_name)      -- get UNO type name

    unoDirectory=.directory~new
    unoNameQueue=.queue~new

    if self~setup=.false then  -- something went wrong, indicate it
    do
       -- drop unoidl_name unoidl_typeName               -- undefine
       unoidl_typeName=.nil            -- indicates that no content could be created
       unoidl_className=.nil
       unoidl_definition=pp(unoidl_name) "could not be successfully processed for this class" pp(self~class~string)
       return
    end


::method supplier    -- the supplier will show the enum-names in queue-order (starting out with the default value)
  expose unoDirectory unoNameQueue  -- enumRelation2NumericValues
  indexes=.array~new
  values =.array~new
  i=0
  do key over unoNameQueue
     i=i+1
     indexes[i]=key
     values[i] =unoDirectory~entry(key)
  end
  return .supplier~new(values,indexes)


::method makearray   -- the makearray will show the enum-names in queue-order (starting out with the default value)
  expose unoNameQueue   -- enumRelation2NumericValues
  indexes=.array~new
  do i=1 to unoNameQueue~items
     indexes[i]=unoNameQueue[i]
  end
  return indexes



::method "[]"
  forward message ("ENTRY")   -- let ENTRY method carry out the work


::method "AT"
  forward message ("ENTRY")   -- let ENTRY method carry out the work


::method entry                -- directory-like method
  expose unoDirectory
  parse arg index             -- make sure we have a string value
  forward to (unoDirectory) array (index)   -- let the directory object carry out the work


::method hasentry             -- directory-like method
  expose unoDirectory
  forward to (unoDirectory)   -- let the directory object carry out the work


::method unknown              -- directory-like method
  expose unoDirectory
  forward to (unoDirectory)   -- let the directory object carry out the work



/* USE BSF-routine "bsf.wrapStaticFields(strJavaClassName)" instead (available as of 2006-02-03) */

/* *************************************************************
   Class:      UNO_ENUM
   Purpose:    make all UNO_ENUMS available via an ooRexx directory object
   Remark:     - there will be an additional entry "DEFAULT_VALUE" created
               - Java-UNO expects enumeration values to be Java objects, which
                 are stored with the constant fields of the Java class representing
                 the UNOIDL-ENUM; this implementation is able to return the enum field name
                 representing the Java enum object;
               - sending the enum name or its numeric value will return the
                 appropriate Java-enum-object

   ************************************************************* */
::CLASS UNO_ENUM SUBCLASS UNO_DIRLIKE PUBLIC

/* Sets up the directory entries for this UNO_ENUM, will be invoked via UNO_DIRLIKE's constructor. */
::method setup       -- check and setup directory object
  if self~unoidl_typeName<>"UNO_ENUM" then          -- of correct UNO type ?
     return .false

  unoidl_name=self~unoidl_name
  self~unoidl_definition=uno.getDefinition(unoidl_name)  -- get the UNOIDL-definition

  enumDirectory=self~unoDirectory
  enumQueue    =self~unoNameQueue

       -- extract and save the (Java) className
  parse value self~unoidl_definition with . "|" className "|" .
  self~unoidl_className=className

  tmpArr=self~unoidl_definition~makearray(" ")  -- split string at blank character

  parse value tmpArr[1] with . "|" . "|" defVal rest   -- ENUM: default value, CONSTANTS: typeClass

  enumDirectory~default_Value=defVal    -- save default value
  enumQueue~queue("DEFAULT_VALUE")

  do i=2 to tmpArr~items                        -- iterate over ENUM values
     parse value tmpArr[i] with name "|" val
       -- with UNO_ENUM use the Java object which is created by the Java UNO language binding
       -- and *not* the defined integer value
     enumDirectory~setentry(bsf.getConstant(unoidl_name, name)~string, name)
     enumDirectory~setentry(val,  name)         -- save numeric value in directory
     enumDirectory~setentry(name, val)          -- save numeric value in directory
     enumQueue~queue(name)                      -- queue enum-name
  end
  return .true                                  -- everything worked out o.k.





/* USE BSF-routine "bsf.wrapStaticFields(strJavaClassName)" instead (available as of 2006-02-03) */

/* *************************************************************
   Class:      UNO_CONSTANTS
   Purpose:    make all UNO_CONSTANTS name/value pairs available via an ooRexx directory object
   Remark:     - sending the constant name will return its (numeric) value
   ************************************************************* */
::CLASS UNO_CONSTANTS SUBCLASS UNO_DIRLIKE PUBLIC
::method setup       -- check and setup directory object

  if self~unoidl_typeName<>"UNO_CONSTANTS" then     -- of correct UNO type ?
    return .false

  unoidl_name=self~unoidl_name
  self~unoidl_definition=uno.getDefinition(unoidl_name)  -- get the UNOIDL-definition

  enumDirectory=self~unoDirectory
  enumQueue    =self~unoNameQueue

       -- extract and save the (Java) className
  parse value self~unoidl_definition with . "|" className "|" .
  self~unoidl_className=className

  tmpArr=self~unoidl_definition~makearray(" ")

  do i=2 to tmpArr~items                        -- iterate over CONSTANT values
     parse upper value tmpArr[i] with name "|" val
     enumDirectory[name]=val
     enumDirectory[val] =name
     enumQueue~queue(name)                      -- queue enum-name
  end
  return .true                                  -- everything worked out o.k.


/* Decodes the received number (a result of adding up constant values of this class)
   into a string containing the constant field names that are set in the received number.
*/
::method decode
  parse arg number         -- retrieve the number, can consists of added constants
  maxDigits=20
  numeric digits maxDigits -- make sure that we have enough digits to exactly represent 2**64+1

     -- turn decimal number into hexadecimal value
  if number<0 then
     tmpVal=number~d2c(maxDigits)
  else
     tmpVal=number~d2c

  lenTV=length(tmpVal)     -- length of number

  tmpStr=""
  do key over self         -- iterate over all keys
     tmp=self~entry(key)               -- get decimal value for key
     if tmp<0 then tmpKey=tmp~d2c(maxDigits) -- turn decimal value into character
              else tmpKey=tmp~d2c

     lenTK=length(tmpKey)  -- lenght of key
     if lenTK>lenTV then   -- constant bigger than available number, leave
        leave

     if bitand(right(tmpVal, lenTK), tmpKey)=tmpKey then
        tmpStr=tmpStr key
  end
  return tmpStr~strip      -- remove leading/trailing blanks


/* Encodes given blank delimited list of constant field names into appropriate decimal number.
   Optionally, one can intermix the list of constant field names with decimal numbers (standing
   for a constant).
*/
::method encode
  parse arg list
  list=changestr("+", list, " ")       -- if plus-signs are supplied, replace them with blanks
  numeric digits 20
  res=0
  do while list<>""                    -- process blank-delimited list
     parse var list key list           -- extract constant field name/decimal value
     if datatype(key, "Numeric") then  -- if numeric, use value to add
        res=res+key
     else
        res=res+self~entry(key)
  end
  return res                           -- return resulting value



/* *************************************************************
   Class:      UNO_PROPERTIES
   Purpose:    make all UNO_PROPERTY definitions via an ooRexx directory object
   Remark:     - sending the property name will return its encoded UNOIDL string, which
                 needs to be parsed for further processing
   ************************************************************* */
::CLASS UNO_PROPERTIES SUBCLASS UNO_DIRLIKE -- PUBLIC
::method setup       -- check and setup directory object

    if self~unoidl_typeName<>"" then                  -- of correct UNO type ?
       return .false

    encodedProps=self~unoidl_name                     -- get argument
    parse var encodedProps . "|" type "|" .           -- check for encoded properties
    if type<>"UNO_PROPERTY" then
       return .false

    self~unoidl_name=""                               -- indicate no UNOIDL name available
    self~unoidl_typeName="UNO_PROPERTY"               -- indicate kind of directory
    self~unoidl_className=""                          -- indicate kind of directory

    self~unoidl_definition=encodedProps               -- save encoded property string

    tmpArr=encodedProps~makearray(" ")

    do i=1 to tmpArr~items                            -- iterate over CONSTANT values
       parse value tmpArr[i] with name "|" .
       self~setentry(name, tmpArr[i])                 -- save encoded property info
    end
    return .true                                      -- everything worked out o.k.



  -- wrap argument into an UNO-proxy object, if it is a BSF object, else return argument unchanged
::routine UNO.WRAP public
  if .uno~proxyclasses~hasindex(arg(1)~class) then    -- already wrapped up
     return arg(1)

  if arg(1)~hasmethod("BSF.EXIT") then
  do
     if abbrev(arg(1)~class~id, "BSF_ARRAY") then
        return uno.wrapArray(arg(1))

     return .UNO_proxy~new(arg(1))        -- wrap it up as an UNO_PROXY
  end

  o=bsf.wrap(arg(1))    -- hmm, maybe not yet wrapped up as a BSF-object?
  if o~hasmethod("BSF.EXIT") then
  do
     if abbrev(o~class~id, "BSF_ARRAY") then
        return uno.wrapArray(o)

     return .UNO_proxy~new(o) -- o.k. a BSF object
  end

  return arg(1)                                 -- return argument as is (unwrapped)


/* Routine that determines whether a BSF array should be wrapped up as an UNO array object,
   which will cause any object that gets accessed from the array via UNO to be wrapped up
   itself as an UNO_PROXY object (allowing to send the name of interfaces to retrieve the
   desired interface object).

   It will not wrap up BSF arrays that have either primitive datatypes as their component
   type, or classes that start with "java" (which are regarded to belong to Java not to UNO).
*/
::routine UNO.WRAPARRAY
  use arg bsfObject

  parse value bsfObject~string with leadin "@" .   -- get lead-in
  pos=verify(leadin , "[")
  if pos>0 then   -- o.k. array symbols are there
  do
     if length(leadin)<>pos then   -- not a primitive dataype in hand
     do
        if abbrev(substr(leadin, pos), "Ljava")=.false then    -- maybe a UNO class, so wrap it up
        do
           return .uno_array_proxy~new(bsfObject)  -- wrap it up as an UNO array object
        end
     end
  end

  return bsfObject



   -- convert a file-name with path to URL notation as needed by UNO
::routine ConvertToUrl public    -- named after {Star|OOo}Basic
  parse arg str
  return (iif(left(str,1)="/", "file://", "file:///") || encodeURL(str))~changestr("\", "/")


   -- convert a file-name with path encoded in the URL notation to an operating system compliant file with path
::routine ConvertFromUrl public  -- named after {Star|OOo}Basic
  parse arg str

  new=decodeURL(str)             -- decode URL (replace escapes with the escaped chars themselves


  if left(new,5)="file:" then
  do
     dosLike= (substr(new, 10,1)=":")  -- if 10th char a colon, assume DOS-like filename

     if dosLike=.true then
        new=substr(new, 9)
     else      -- leave initial slash
        new=substr(new, 8)

     if .UNO~hasentry("FILE.SEPARATOR")=.false then   -- not yet in .UNO directory, hence add it
        .UNO~file.separator=.java.lang.system~getProperty("file.separator")   -- get info from Java

      -- if a DOS-like path and on a DOS-like system, translate slashes to backslashes
     if dosLike=.true | .UNO~file.separator="\" then
        return changestr("/", new, "\")
  end
  return new



::routine encodeURL public
  parse arg str

  new=changestr("%", str, "%25")    -- escape escape char %

   -- determine chars in URLs that do not get encoded
  url.chars="abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789/\?:=%'"".,+-"

  pos=verify(new, url.chars)     -- chars to escape in string?
  do while pos<>0
     char=substr(new, pos, 1)    -- extract char not in reference
     new=new~changestr(char, "%"||char~c2x)  -- escape character
     pos=verify(new, url.chars)
  end
  return new




   -- re-create chars from escape-sequences led in by % followed by two-digit hex value
::routine decodeUrl public    -- %20 --> " "
  parse arg str

  new=""
  do while str<>""
     parse var str before "%" +1 hexValue +2 str -- extract hex value
     new=new || before || hexValue~x2c
  end
  return new




   -- interface to UnoRuntime.areSame(o1, o2)
::routine uno.areSame public
   use arg o1, o2
   return .uno~UnoRuntime~areSame(o1, o2)     /* true if both reference the same UNO object   */


/* Pass-through routines for "org.oorexx.uno.RgfReflectUNO", wrapping up
   returned UNO objects into UNO_PROXY ooRexx objects */
   -- interface to RgfReflectUNO.findInterfaceWithMember(): can be a blank delimited string or an interface object
::routine uno.findInterfaceWithMember public
   use arg o, name, bREturnString, howMany

   if datatype(bReturnString, "Boolean")<>.true then
      bReturnString=.true                       -- default to return a blank delimited string

   if datatype(howMany, "Whole")<>.true then
      howMany=-1                                -- default to return all matches

      -- wrap up as an UNO proxy, if an object got returned
   return uno.wrap( .uno~rgfReflectUNO~findInterfaceWithMember( o, name, bReturnString, howMany, .uno~bExtendSearch) )


   -- interface to RgfReflectUNO.getDefinition(): returns always a blank delimited string
::routine uno.getDefinition public
   use arg o
   return .uno~rgfReflectUNO~getDefinition(o)


   -- interface to RgfReflectUNO.getInterfacesViaReflection(): returns always a blank delimited string
::routine uno.getInterfaceNamesViaReflection public
   use arg o
   return .uno~rgfReflectUNO~getInterfaceNamesViaReflection(o)


   -- interface to RgfReflectUNO.getProperties(): returns always a blank delimited string
::routine uno.getProperties public
   use arg o
   return .uno~rgfReflectUNO~getProperties(o)


   -- interface to RgfReflectUNO.getServicesViaReflection(): returns always a blank delimited string
::routine uno.getServiceNamesViaReflection public
   use arg o
   return .uno~rgfReflectUNO~getServiceNamesViaReflection(o)


   -- interface to RgfReflectUNO.getType(): always a string
::routine uno.getTypeName public
   use arg o
   return .uno~rgfReflectUNO~getTypeName(o)


   -- interface to RgfReflectUNO.getXTypeProviderTypes(): returns always a blank delimited string
::routine uno.getXTypeProviderTypeNames public
   use arg o
   return .uno~rgfReflectUNO~getXTypeProviderTypeNames(o)


   -- interface to RgfReflectUNO.queryInterfaceName(): always a string
::routine uno.queryInterfaceName public
   use arg o, name
   return .uno~rgfReflectUNO~queryInterfaceName(o, name, .uno~bExtendSearch)-- always a string


   -- interface to RgfReflectUNO.queryInterfaceObjectByName(): returns the interface object or .nil
::routine uno.queryInterfaceObjectByName public
   use arg o, name
      -- wrap up as an UNO proxy, if an object got returned
   return uno.wrap( .uno~rgfReflectUNO~queryInterfaceObjectByName(o, name, .uno~bExtendSearch) )


   -- interface to RgfReflectUNO.queryServiceName(): always a string
::routine uno.queryServiceName public
   use arg o, name
   return .uno~rgfReflectUNO~queryServiceName(o, name)


   -- "pretty print" strings from RgfReflectUNO: uses a blank to create an array from the string
   --                first entry is untoched, all others are delimited with CR-LF-TAB for indentation
::routine ppd public
  parse arg debugString, delimiter

  if delimiter=="" then delimiter=" "  -- default to a blank as delimiter

  arr=debugString~makearray(delimiter)
  tab="09"x
  crlf="0d0a"x
  crlftab=crlf||tab
  tmpStr=arr[1]
  do i=2 to arr~items
     tmpStr=tmpStr || crlftab || arr[i]
  end

  return tmpStr



::routine uno.getScriptContext public     -- will be called from an ooRexx macro
scriptContext=uno.wrap(bsf.lookupBean("OOo.xScriptContext"))

if .nil<>scriptContext then   -- make sure RgfReflectUNO has XComponentContext set
do                            -- use the XComponentContext received by the scripting framework
   .uno~rgfReflectUNO~setContext(scriptContext~getComponentContext)
end

return scriptContext



::routine uno.getScriptContextVersion public
return bsf.lookupBean("OOo.ooRexxScriptProvider.version")


/* Create a BSF-array and wrap it up as an UNO_ARRAY_PROXY adding the
   behaviour of wrapping up its elements as UNO_PROXY elements, such that
   messages starting with an "X" are interpreted as ones which query
   an interface object.
*/
::routine uno.createArray              public
      -- create and send message to .BSF class supplying the received arguments
   res=.message~new(.bsf, "BSF.CREATEARRAY", "ARRAY", arg(1, "ARRAY"))~send
   if abbrev(res~class~id, "BSF_ARRAY") then
      return uno.wrapArray(res)

   return res


/* Returns the fully qualified, mixed case name of the interface. Either pass in
   the case-independent unqualified or fully qualified name of the interface.

   If an unqualified interface name matches duplicates they are returned, delimited
   with the char(acter sequence) as denominated by the second argument. If the second
   argument is missing, then a single space character (" ") is used instead.

*/
::routine uno.getInterfaceName      public   --rgf, 2006-09-03
  parse upper arg name, needle

  tmpStr=.uno~XInterfaces~entry(name)  -- unknown interface
  if .nil=tmpStr then return tmpStr

  bAddNeedle=.false
  if ARG(2, "O") then      -- Omitted, then default to CR-LF
  do
     needle=" "         -- default to single space
     -- needle="0d0a"x
  end

  do entry over .UNO~XInterfaces.dupes~allAt(name)
     tmpStr=tmpStr || needle || entry
  end
  return tmpStr

/*


---
  a=.uno~xinterfaces
  .UNO~XInterfaces

   if xClass=.nil then        -- no entry as of yet
   do
      if xName<>idx then      -- long name given, then use that
      do
         xClass=UNO.loadClass(.UNO~XInterfaces~entry(xName), xName)
         if xClass<>.nil then -- o.k. also save with class name
         do
             .UNO~setEntry(idx, xClass)
         end
      end
      else     -- only class name is given
      do
         xClass=UNO.loadClass(.UNO~XInterfaces~entry(idx), idx)
         if xClass<>.nil then -- o.k., also save with full class name
         do
             .UNO~setEntry(.UNO~XInterfaces~entry(idx), xClass)
         end
      end
   end

   return xclass

---
   if idx~left(1)="X" then    -- if an interface object is asked for, query and return it
   do
      xClass=getInterfaceClass(xName, idx)

         -- query and return the interface object
      interface=.UNO~unoRuntime~queryInterface(xClass, bsfObject)
      if interface=.nil then  -- not found, maybe a duplicate simple class name ?
      do
         do fullName over .UNO~XInterfaces.dupes~allAt(idx)
            idx=substr(fullName, lastpos(".", fullName)+1)   -- getindex
            xClass=getInterfaceClass(fullName, idx)      -- rgf, 2006-06-04

            if .nil<>xClass then
               interface=.UNO~unoRuntime~queryInterface(xClass, bsfObject)

            if .nil<>interface then
               return .UNO_proxy~new(interface)
         end

         -- now try reflection (via "RgfReflectUNO"), maybe a new interface, not in UNO_XINTERFACES.REX
         res=.uno~RgfReflectUno~queryInterfaceObjectByName(bsfObject, xName, .uno~bExtendSearch )

         if .nil<>res then
            return .uno_proxy~new(res)

         return .nil    -- no interface object available
      end

      return .UNO_proxy~new(interface)
   end
*/
